
#include <stdio.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gconf/gconf-client.h>
#include <gconf/gconf-value.h>
#include <libp11.h>

#include "panel-dbus-service.h"
#include "panel-session-manager.h"

static GMainLoop *loop;
static GConfClient *gconf;
static PanelSessionManager *session;

volatile guint poll_time;
volatile guint logout_time;
volatile guint poll_timeout = 0;
volatile guint logout_timeout = 0;

gchar *pkcs11_module;
volatile PKCS11_CTX *pkcs11_ctx;
volatile PKCS11_SLOT *pkcs11_slots;
volatile int pkcs11_nslots;

volatile gboolean module_loaded;
volatile gboolean card_present;

void pkcs11_free_slots()
{
	if (pkcs11_ctx && pkcs11_slots) {
		PKCS11_release_all_slots(pkcs11_ctx, pkcs11_slots, pkcs11_nslots);
		pkcs11_slots = NULL;
		pkcs11_nslots = 0;
	}
}

int pkcs11_get_token()
{
	pkcs11_free_slots();

	if (PKCS11_enumerate_slots(pkcs11_ctx, &pkcs11_slots, &pkcs11_nslots) ||
		pkcs11_slots == NULL || pkcs11_nslots == 0) {
		pkcs11_free_slots();
		return -1;
	}

	if (PKCS11_find_token(pkcs11_ctx, pkcs11_slots, pkcs11_nslots) == NULL)
		return 0;

	return 1;
}

gboolean pkcs11_logout(gpointer data)
{
	panel_session_manager_request_logout(session, PANEL_SESSION_MANAGER_LOGOUT_MODE_FORCE);

	g_main_loop_quit(loop);
	return FALSE;
}

void pkcs11_lock()
{
	if (logout_timeout)
		g_source_remove(logout_timeout);
	logout_timeout = logout_time > 0 ? g_timeout_add(logout_time * 1000, pkcs11_logout, NULL) : 0;

	panel_lock_screen_action(gdk_screen_get_default(), "lock");
}

void pkcs11_unlock()
{
	if (logout_timeout)
		g_source_remove(logout_timeout);
	logout_timeout = 0;

	panel_lock_screen_action(gdk_screen_get_default(), "deactivate");
}

gboolean pkcs11_poll(gpointer data)
{
	int token_state;

	token_state = pkcs11_get_token();

	switch((card_present ? 1000 : 0) + token_state) {
	case    0:
		/* not present, not available -> no change */
		break;
	case 1001:
		/* card present, still available -> no change */
		break;
	case 1000:
		/* card present, not available -> removed */
		fprintf(stderr, "Card removed.\n");
		PKCS11_CTX_unload(pkcs11_ctx);
		module_loaded = FALSE;
		if (PKCS11_CTX_load(pkcs11_ctx, pkcs11_module)) {
			g_main_loop_quit(loop);
			return FALSE;
		}
		module_loaded = TRUE;
		card_present = FALSE;

		pkcs11_lock();
		break;
	case    1:
		/* not present, available -> inserted */
		fprintf(stderr, "Card inserted.\n");
		card_present = TRUE;

		pkcs11_unlock();
		break;
	default:
		/* error */
		g_main_loop_quit(loop);
		return FALSE;
	}

	return TRUE;
}

void pkcs11_gconf_notify(GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
	const char *key;
	const GConfValue *value;

	if (entry == NULL || (key = gconf_entry_get_key(entry)) == NULL)
		return;

	if (!strcmp(key, "/apps/pkcs11-eventmgr/poll_time")) {
		if ((value = gconf_entry_get_value(entry)) == NULL)
			return;

		poll_time = gconf_value_get_int(value);
		if (poll_timeout)
			g_source_remove(poll_timeout);
		poll_timeout = g_timeout_add(poll_time, pkcs11_poll, NULL);
	} else if (!strcmp(key, "/apps/pkcs11-eventmgr/logout_time")) {
		logout_time = gconf_value_get_int(value);
	}

	return;
}

int main(int argc, char *argv[])
{
	int ret=1;

	g_type_init();
	gdk_init(&argc, &argv);

	if ((gconf = gconf_client_get_default()) == NULL) {
		fprintf(stderr, "GConf\n");
		goto out;
	}

	if ((session = panel_session_manager_get()) == NULL) {
		fprintf(stderr, "Session\n");
		goto out;
	}

	if ((loop = g_main_loop_new(NULL, FALSE)) == NULL) {
		fprintf(stderr, "MainLoop\n");
		goto out;
	}

	poll_time = gconf_client_get_int(gconf, "/apps/pkcs11-eventmgr/poll_time", NULL);
	logout_time = gconf_client_get_int(gconf, "/apps/pkcs11-eventmgr/logout_time", NULL);
	pkcs11_module = gconf_client_get_string(gconf, "/apps/pkcs11-eventmgr/module", NULL);

	if (poll_time == 0)
		poll_time = 500;

	fprintf(stderr, "Startup with %d/%d/\"%s\"\n", poll_time, logout_time, pkcs11_module);

	if ((pkcs11_ctx = PKCS11_CTX_new()) == NULL) {
		fprintf(stderr, "PKCS#11 CTX\n");
		goto out_mainloop;
	}

	module_loaded = FALSE;
	if (PKCS11_CTX_load(pkcs11_ctx, pkcs11_module)) {
		fprintf(stderr, "PKCS#11 Load\n");
		goto out_pkcs11_ctx;
	}
	module_loaded = TRUE;

	if (PKCS11_enumerate_slots(pkcs11_ctx, &pkcs11_slots, &pkcs11_nslots) ||
		pkcs11_slots == 0 || pkcs11_nslots == 0) {
		fprintf(stderr, "PKCS#11 Slots\n");
		goto out_pkcs11_module;
	}

	if (PKCS11_find_token(pkcs11_ctx, pkcs11_slots, pkcs11_nslots) == NULL) {
		fprintf(stderr, "PKCS#11 Token\n");
		goto out_pkcs11_slots;
	}

	card_present = TRUE;

	poll_timeout = g_timeout_add(poll_time, pkcs11_poll, NULL);
	gconf_client_notify_add(gconf, "/apps/pkcs11-eventmgr", pkcs11_gconf_notify, NULL, NULL, NULL);

	g_main_loop_run(loop);
	ret = 0;

out_pkcs11_slots:
	pkcs11_free_slots();
out_pkcs11_module:
	if (module_loaded)
		PKCS11_CTX_unload(pkcs11_ctx);
out_pkcs11_ctx:
	PKCS11_CTX_free(pkcs11_ctx);
out_mainloop:
	g_main_loop_unref(loop);
out:
	fprintf(stderr, "exit(%d)\n", ret);
	exit(ret);
}
